import datetime
import json
import os
from .utils import create_safe_file_name

# generator method to make websocket creation easier
# keeps us to standard format
# we check passed in params to know what pipes to look for and what data to expect
def create_gen_txt_websocket(sock, plugin_name, pipe_dict, default_params = {}, forward_pass=None):
    def generate_txt_common(ws):
        while True:
            data_str = ws.receive()
            if not data_str:
                continue
            data = json.loads(data_str)
            prompt = data.get('prompt')
            max_tokens = data.get('max_tokens') or 512
            temperature = data.get('temperature') or 0
            if not prompt:
                ws.send(json.dumps({'error': 'Prompt is required'}))
                continue
            
            model = pipe_dict["pipeline"]

            response = ""
            token_count = 0
            tokens = []
            for token in model.generate(prompt, max_tokens=max_tokens, temp=temperature, streaming=True):
                tokens.append(token)
                token_count += 1
                response += token
                ws.send(json.dumps({'text': token, "count": token_count}))

            # Save the result to a local file
            # Use the date and prompt to name the file
            date_str = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
            file_name = f"{date_str}_{prompt.replace(' ', '_')}.txt"
            inference_directory = os.getenv("INFERENCE_DIRECTORY", "./inference")
            out_path = os.path.join(inference_directory, file_name)
            out_dir = os.path.dirname(out_path)
            
            # Make the directory if it doesn't exist
            if not os.path.exists(out_dir):
                os.makedirs(out_dir)

            safe_file_name = create_safe_file_name(file_name)
            safe_out_path = os.path.join(inference_directory, safe_file_name)

            with open(safe_out_path, 'w') as f:
                f.write(response)
            
            # Return the file path as a response
            ws.send(json.dumps({
                "file_path": safe_out_path,
                'text': response
            }))

    @sock.route('/ws/generate/txt')
    def generate_txt(ws):
        generate_txt_common(ws)

    @sock.route('/ws/generate')
    def generate(ws):
        generate_txt_common(ws)


def create_gen_chat_websocket(sock, plugin_name, pipe_dict, default_params = {}, forward_pass=None):
    def generate_chat_common(ws):
         # we can hold the chat session for the entire websocket connection
        # they can disconnect and reconnect, send an id to re-hydrate, or its a new session
        model = pipe_dict["pipeline"]
        with model.chat_session() as chat:
            while True:
                data_str = ws.receive()
                if not data_str:
                    continue
                data = json.loads(data_str)
                prompt = data.get('prompt')
                max_tokens = data.get('max_tokens') or 512
                temperature = data.get('temperature') or 0
                if not prompt:
                    ws.send(json.dumps({'error': 'Prompt is required'}))
                    continue

                historical_messages = data.get('historicalMessages')
                if historical_messages:
                    # waiting on gpt4all to add official reload support, this will leak data (dont use for shared inference)
                    model._history = historical_messages

                response = ""
                token_count = 0
                tokens = []
                for token in model.generate(prompt, max_tokens=max_tokens, temp=temperature, streaming=True):
                    tokens.append(token)
                    token_count += 1
                    response += token
                    ws.send(json.dumps({'text': token, "count": token_count}))

                # Save the result to a local file
                # Use the date and prompt to name the file
                date_str = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
                file_name = f"{date_str}_{prompt.replace(' ', '_')}.txt"
                inference_directory = os.getenv("INFERENCE_DIRECTORY", "./inference")
                out_path = os.path.join(inference_directory, file_name)
                out_dir = os.path.dirname(out_path)
                
                # Make the directory if it doesn't exist
                if not os.path.exists(out_dir):
                    os.makedirs(out_dir)

                # Replace special characters in the file name to avoid OSError
                safe_file_name = create_safe_file_name(file_name)
                safe_out_path = os.path.join(inference_directory, safe_file_name)

                with open(safe_out_path, 'w') as f:
                    f.write(response)
                
                # Return the file path as a response
                ws.send(json.dumps({
                    "file_path": safe_out_path,
                    'text': response,
                    "historicalMessages": model.current_chat_session
                }))


    @sock.route('/ws/generate/chat')
    def generate_chat(ws):
        generate_chat_common(ws)

    @sock.route('/ws/chat')
    def chat(ws):
        generate_chat_common(ws)



import asyncio
import datetime
import json
import os
import shutil
import signal
import time
import click
import requests
import logging
import psutil
import tqdm
from tenacity import retry, stop_after_attempt, wait_exponential

from cli.fingerprint import get_fingerprint
from cli.lib import create_venv, get_and_download_plugin, get_current_user, get_inference_job, get_my_workers, install_requirements, install_torch_requirements, kill_process_on_port, login_with_token, post_register_worker, process_inference_job, run_plugin

logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')
logger = logging.getLogger(__name__)

@click.group()
def cli():
    pass

@click.command(help="Share my AI CLI - Help")
def help():
    click.echo(f'Help:')
    click.echo(f'  - help: Show this help message')
    return

@click.command(help="Login with a token")
@click.option('--token', required=True, help='Your login token')
def login(token):
    logged_in_user = login_with_token(token)
    if not logged_in_user:
        click.echo('Login failed, please try again.')
        return
    click.echo(f'Logged in successfully as {logged_in_user["name"]} - {logged_in_user["id"]}')
    return

@click.command(help="Get the current user")
def current_user():
    user = get_current_user()
    formatted_json = json.dumps(user, indent=4, sort_keys=True)
    click.echo(f'Current user:\n{formatted_json}')
    return

@click.command(help="Get the current machines fingerprint")
def fingerprint():
    fingerprint_bytes = asyncio.run(get_fingerprint())
    fingerprint_str = fingerprint_bytes.hex()
    click.echo(f'Fingerprint: {fingerprint_str}')
    return

@click.command(help="Register new worker")
@click.option('--plugin_id', required=True, help='The plugin id to work on')
def register_worker(plugin_id):
    worker = post_register_worker(plugin_id)
    formatted_json = json.dumps(worker, indent=4, sort_keys=True)
    click.echo(f'Worker:\n{formatted_json}')
    return

@click.command(help="Get my workers")
def my_workers():
    workers = get_my_workers()
    formatted_json = json.dumps(workers, indent=4, sort_keys=True)
    click.echo(f'Workers:\n{formatted_json}')
    return

@click.command(help="Run worker")
@click.option('--plugin_id', required=True, help='The plugin id to work on')
def run_worker(plugin_id):
    # first, register a worker
    worker = post_register_worker(plugin_id)
    formatted_json = json.dumps(worker, indent=4, sort_keys=True)
    click.echo(f'Worker:\n{formatted_json}')
    # then, download the plugin
    plugin, plugin_path = get_and_download_plugin(plugin_id)
    click.echo(f'Plugin downloaded to {plugin_path}')
    # then, create a venv for the plugin
    venv_path = create_venv(plugin_path)
    click.echo(f'Venv created at {venv_path}')
    # then, install the plugin requirements
    install_torch_requirements(venv_path, plugin['torchRequirements'])
    install_requirements(venv_path, plugin['requirements'])
    click.echo(f'Requirements installed')
    # then, run the plugin
    pid = run_plugin(venv_path, plugin_path, plugin['name'])
    try:
        while True:
            with click.progressbar(length=0, label='Checking for jobs', show_eta=False, show_pos=True) as bar:
                while True:
                    job = get_inference_job(worker['id'])
                    if not job:
                        bar.update(0)
                        bar.label = f'No job found - {datetime.datetime.now()}...'
                        time.sleep(5)
                        continue
                    bar.label = f'Job found: {job}'
                    break
            # process it 
            process_inference_job(
                job['params'].get('prompt', ''),
                job['params'].get('image', ''),
                job['params'],
                plugin['params'],
                job['id']
            )
            time.sleep(5)
    except KeyboardInterrupt:
        click.echo(f'Stopping plugin - {datetime.datetime.now()}...')
        try:
            os.kill(pid, signal.SIGTERM)
        except ProcessLookupError:
            pass
    click.echo(f'Plugin stopped - {datetime.datetime.now()}...')
    return

PLUGIN_BASE_DIR = './plugin_cache'

@click.command(help="Run worker that will dynamically download and run plugins")
@click.option('--reset', default=False, help='Reset the worker')
def run_multiplugin_worker(reset):
    plugin_processes = {}

    # if we are resetting, delete all our downloaded plugins 
    if reset:
        for item in os.listdir(PLUGIN_BASE_DIR):
            item_path = os.path.join(PLUGIN_BASE_DIR, item)
            if os.path.isfile(item_path):
                os.remove(item_path)
            elif os.path.isdir(item_path):
                shutil.rmtree(item_path)
        # delete plugin cache file if it exists
        if os.path.exists("cached_plugins.json"):
            os.remove("cached_plugins.json")
        # stop all plugins 
        kill_process_on_port(3030)
        click.echo(f'Deleted all downloaded plugins')
        return

    # first, register a worker
    worker = post_register_worker()
    formatted_json = json.dumps(worker, indent=4, sort_keys=True)
    click.echo(f'Worker:\n{formatted_json}')

    cached_plugins_file = "cached_plugins.json"
    if os.path.exists(cached_plugins_file):
        with open(cached_plugins_file, "r") as f:
            cached_plugins = json.load(f)
    else:
        cached_plugins = {"current": ""}

    while True:
        try:
            logger.info(f'Checking for jobs - {datetime.datetime.now()}...')

            @retry(stop=stop_after_attempt(3), wait=wait_exponential(multiplier=1, min=4, max=10))
            def get_job():
                return get_inference_job(worker['id'])

            job = get_job()

            if not job:
                logger.info(f'No job found - {datetime.datetime.now()}...')
                time.sleep(5)
                continue

            logger.info(f'Job found: {job}')

            # get the plugin id from the job
            plugin_id = job['pluginId']

            # see if its in our cache, if it is, get its name 
            if plugin_id in cached_plugins:
                current_plugin_name = cached_plugins[plugin_id]['plugin']['name']
            else:
                current_plugin_name = ""

            is_current_plugin = False
            if current_plugin_name != "":
                # we can get our plugins name from the :3030/ route 
                try:
                    response = requests.get(f"http://localhost:3030/")
                    # the name is in the string, lowercase and compare 
                    is_current_plugin = lambda plugin_name: response.text.lower().includes(current_plugin_name.lower())
                except Exception as e:
                    pass

            if not is_current_plugin:
                # Kill the current running plugins 
                kill_process_on_port(3030)

                if plugin_id in cached_plugins:
                    # if the plugin is already cached, use the cached version
                    plugin = cached_plugins[plugin_id]['plugin']
                    plugin_path = cached_plugins[plugin_id]['plugin_path']
                    venv_path = cached_plugins[plugin_id]['venv_path']
                    click.echo(f'Using cached plugin: {plugin_id}')
                else:
                    # if the plugin is not cached, download and set it up
                    plugin, plugin_path = get_and_download_plugin(plugin_id, PLUGIN_BASE_DIR)
                    click.echo(f'Plugin downloaded to {plugin_path}')

                    venv_path = create_venv(plugin_path)
                    click.echo(f'Venv created at {venv_path}')

                    install_torch_requirements(venv_path, plugin['torchRequirements'])
                    install_requirements(venv_path, plugin['requirements'])
                    click.echo(f'Requirements installed')

                    # cache the plugin for future use
                    cached_plugins[plugin_id] = {
                        'plugin': plugin,
                        'plugin_path': plugin_path,
                        'venv_path': venv_path
                    }
                    # save the updated cached_plugins to file
                    with open(cached_plugins_file, "w") as f:
                        json.dump(cached_plugins, f)
                    
                pid = run_plugin(venv_path, plugin_path, plugin['name'])
                plugin_processes[plugin_id] = pid
            else:
                venv_path = cached_plugins[plugin_id]['venv_path']
                plugin_path = cached_plugins[plugin_id]['plugin_path']
                plugin = cached_plugins[plugin_id]['plugin']
                
            # Run the plugin only if it's not already running
            #if plugin_id not in plugin_processes or not psutil.pid_exists(plugin_processes[plugin_id]):
            # Kill any existing process with the same name or port

            # wait till we get a response from our healthcheck endpoint
            while True:
                print('Waiting for healthcheck...')
                try:
                    response = requests.get(f"http://localhost:3030/healthcheck")
                except Exception as e:
                    print('Healthcheck failed, retrying in 5 seconds...')
                    time.sleep(5)
                    continue
                if response.status_code == 200:
                    print('Healthcheck passed, continuing...')
                    break
                print('Healthcheck failed, retrying in 5 seconds...')
                time.sleep(5)

            with tqdm.tqdm(desc=f'Processing job {job["id"]}', unit='step') as pbar:
                @retry(stop=stop_after_attempt(3), wait=wait_exponential(multiplier=1, min=4, max=10))
                def process_job():
                    process_inference_job(
                        job['params'].get('prompt', ''),
                        job['params'].get('image', ''),
                        job['params'],
                        plugin['params'],
                        job['id']
                    )
                    pbar.update(1)

                process_job()

        except KeyboardInterrupt:
            logger.info(f'Stopping worker - {datetime.datetime.now()}...')
            break
        except Exception as e:
            logger.exception(f'Error occurred: {str(e)}')
            time.sleep(5)
            
    return

cli.add_command(help)
cli.add_command(login)
cli.add_command(current_user)
cli.add_command(fingerprint)
cli.add_command(register_worker)
cli.add_command(my_workers)
cli.add_command(run_worker)
cli.add_command(run_multiplugin_worker)






import base64
import datetime
import gc
import io
import json
import os
import sys
from PIL import Image
from .utils import save_gif

def create_gen_video_websocket(sock, pipe_dict, default_params = {}, forward_pass=None):
    def generate_video_common(ws):
        while True:
            print("Waiting for data")
            data_str = ws.receive()
            if not data_str:
                print("No data received, skipping...")
                continue
            print(f"Data received: {len(data_str)} bytes")
            try:
                data = json.loads(data_str)
            except json.JSONDecodeError as e:
                print(f"Failed to decode JSON: {str(e)}")
                ws.send(json.dumps({'error': 'Invalid JSON data'}))
                continue
            params = {**default_params, **data}

            if not params['prompt'] and not params['image']:
                ws.send(json.dumps({'error': 'Either prompt or image is required'}))
                continue

            pipe = pipe_dict['pipeline']
            
            # Check if the image is sent as base64 or as a file object
            if 'image' in params and params['image']:
                image = None
                # First, check if it looks like a file path or object
                if any(ext in params['image'] for ext in ['.jpg', '.jpeg', '.png', '.gif']):
                    try:
                        image = Image.open(params['image'])
                    except FileNotFoundError:
                        ws.send(json.dumps({'error': 'Image file not found'}))
                    except Exception as e:
                        ws.send(json.dumps({'error': 'Error opening image file'}))
                else:
                    # If it doesn't look like a file, attempt to decode as base64
                    try:
                        image_data = base64.b64decode(params['image'].split(',')[-1])  # Handle potential data URI scheme
                        image = Image.open(io.BytesIO(image_data))
                    except Exception as e:
                        ws.send(json.dumps({'error': 'Invalid or unrecognized image data'}))
            else:
                image = None

            def callback(_, step_index, timestep, kwargs):
                step = step_index + 1
                ws.send(json.dumps({
                    'step': step,
                    'steps': params.get('inference_steps', params.get('num_frames')),
                    'progress': f'{(step/params.get("inference_steps", params.get("num_frames")))*100:.2f}%'
                }))
                return kwargs
            
            # Auto convert strings to numbers for all params
            for key in params:
                if isinstance(params[key], str) and params[key].isdigit():
                    params[key] = int(params[key])
                elif isinstance(params[key], str) and '.' in params[key]:
                    try:
                        params[key] = float(params[key])
                    except ValueError:
                        pass
            if forward_pass:
                result = forward_pass(pipe, params)
            else:
                if image is None:
                    result = pipe(prompt=params['prompt'], decode_chunk_size=params['decode_chunk_size']).frames[0]
                else:
                    image = image.resize((params['width'], params['height']))
                    print(f"img2img image={image}")
                    print(f"pipe_params={params}")
                    result = pipe(image=image, decode_chunk_size=params['decode_chunk_size'], callback_on_step_end=callback).frames[0]
                
            prompt_or_time = params['prompt'] if params['prompt'] else datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
            truncated_prompt = prompt_or_time[:20]
            out_path = save_gif(result, truncated_prompt, loop=0)

            full_path = os.path.abspath(out_path)

            ws.send(json.dumps({
                'file_path': full_path
            }))

            del result, image, pipe
            if 'torch' in sys.modules:
                import torch
                torch.cuda.empty_cache()
            gc.collect()

    @sock.route('/ws/generate/video')
    def generate_video(ws):
        generate_video_common(ws)

    @sock.route('/ws/generate')
    def generate(ws):
        generate_video_common(ws)


'''
run v1 plugins
'''
import os
import subprocess
import sys
import shutil
import re
import glob
import json
import urllib.request
import zipfile
from pathlib import Path

def download_file(url, savepath):
    """
    Download the file and save to savepath
    """
    print(f'Downloading {url}')
    urllib.request.urlretrieve(url, savepath)

def install_python(python_ver, python_path):
    """
    Download and install python
    """
    URL_PYTHON = f'https://www.python.org/ftp/python/{python_ver}/python-{python_ver}-embed-amd64.zip'
    URL_PIP = r'https://bootstrap.pypa.io/get-pip.py'
    URL_MSVC = r'https://sharemyai-releases-dev.s3.amazonaws.com/msvc.zip'
    python_zip_path = python_path / 'python.zip'
    pip_path = python_path / 'get-pip.py'
    msvc_zip_path = python_path / 'msvc.zip'
    
    # Download files if they do not exist
    if not python_zip_path.exists() and not (python_path / 'python.exe').exists():
        download_file(URL_PYTHON, python_zip_path)
    if not pip_path.exists() and not (python_path / 'Scripts' / 'pip.exe').exists():
        download_file(URL_PIP, pip_path)
    if not msvc_zip_path.exists() and not any(python_path.glob('msvc*.dll')):
        download_file(URL_MSVC, msvc_zip_path)
    
    # Extract python zip if not already extracted
    if not (python_path / 'python.exe').exists():
        print("Extracting python")
        with zipfile.ZipFile(python_zip_path, 'r') as zip_ref:
            zip_ref.extractall(python_path)
        python_zip_path.unlink()
    
    # Remove _pth file if it exists
    for pth_file in python_path.glob("*._pth"):
        pth_file.unlink()
    
    # Install MS VC dlls if not already installed
    if msvc_zip_path.exists() and not (python_path / 'msvc').exists():
        print("Installing MSVC dlls")
        with zipfile.ZipFile(msvc_zip_path, 'r') as zip_ref:
            zip_ref.extractall(python_path)
        msvc_zip_path.unlink()
    
    # Install pip if not already installed
    if not (python_path / 'Scripts' / 'pip.exe').exists():
        pip_process = subprocess.Popen(args='python.exe get-pip.py', cwd=str(python_path), shell=True)
        pip_process.wait()
        pip_path.unlink()
        # No need to kill the process as wait() ensures the process has terminated
        print('pip installation process completed')

# https://astral.sh/blog/uv lets go fast
# causes problems with check_call and piping the output to stdout, will need to cache and tree shake ourselves
def install_uv(python_executable):
    subprocess.check_call([python_executable, "-m", "pip", "install", "uv"], env=set_environment_variables(python_executable))

def install_dependencies(python_executable, dependencies):
    subprocess.check_call([python_executable, "-m", "pip", "install"] + dependencies, env=set_environment_variables(python_executable))

def install_sharemyai(python_executable):
    subprocess.check_call([python_executable, "-m", "pip", "install", "sharemyai_utils", "--upgrade"], env=set_environment_variables(python_executable))

def install_torch_dependencies(python_executable, dependencies):
    print("Python architecture:", "64-bit" if sys.maxsize > 2**32 else "32-bit")
    torch_index_url = "https://download.pytorch.org/whl/cu118"
    print("Installing torch dependencies from", torch_index_url)
    subprocess.check_call([python_executable, "-m", "pip", "install"] + dependencies + ["--index-url", torch_index_url], env=set_environment_variables(python_executable))
    
# todo, this should also use UV
def install_sharemyai_utils(python_executable):
    # temp workaround for sharemyai_utils
    # Check and install specific version of setuptools if version changed
    setuptools_version = "58.2.0"
    try:
        installed_version = subprocess.check_output([python_executable, "-m", "pip", "show", "setuptools"], env=set_environment_variables(python_executable)).decode().split('\n')[1].split(': ')[1].strip()
        if installed_version != setuptools_version:
            print(f"Updating setuptools from version {installed_version} to {setuptools_version}")
            subprocess.check_call([python_executable, "-m", "pip", "install", f"setuptools=={setuptools_version}"], env=set_environment_variables(python_executable))
    except subprocess.CalledProcessError:
        print("setuptools not found, installing version", setuptools_version)
        subprocess.check_call([python_executable, "-m", "pip", "install", f"setuptools=={setuptools_version}"], env=set_environment_variables(python_executable))
    sharemyai_utils_path = os.path.join(os.path.dirname(__file__), "sharemyai_utils")
    sharemyai_utils_dest_path = os.path.join(os.path.dirname(python_executable), "sharemyai_utils")
    
    # Check if sharemyai_utils version has changed before attempting re-installation or copying
    try:
        installed_sharemyai_utils_version = subprocess.check_output([python_executable, "-m", "pip", "show", "sharemyai_utils"], env=set_environment_variables(python_executable)).decode().split('\n')[1].split(': ')[1]
        with open(os.path.join(sharemyai_utils_path, 'sharemyai_utils', '__init__.py'), 'r') as f:
            local_version = f.readline().strip().split('=')[1].replace("'", "").strip()
        installed_sharemyai_utils_version = installed_sharemyai_utils_version.strip()
        local_version = local_version.strip()
        print(f"Installed sharemyai_utils version: {installed_sharemyai_utils_version}, Local sharemyai_utils version: {local_version}")
        if installed_sharemyai_utils_version.strip() == local_version.strip():
            print("Versions match. No update required.")
        else:
            print("Version mismatch detected. Update required.")
            print(f"Updating sharemyai_utils from version {installed_sharemyai_utils_version} to {local_version}")
            print("Copying sharemyai_utils to", sharemyai_utils_dest_path)
            shutil.copytree(sharemyai_utils_path, sharemyai_utils_dest_path, dirs_exist_ok=True)
            subprocess.check_call([python_executable, "-m", "pip", "install", "-e", sharemyai_utils_dest_path], env=set_environment_variables(python_executable))
            print("sharemyai_utils successfully installed or updated.")
    except subprocess.CalledProcessError as e:
        print("Installing sharemyai_utils for the first time.")
        print("Copying sharemyai_utils to", sharemyai_utils_dest_path)
        shutil.copytree(sharemyai_utils_path, sharemyai_utils_dest_path, dirs_exist_ok=True)
        subprocess.check_call([python_executable, "-m", "pip", "install", "-e", sharemyai_utils_dest_path], env=set_environment_variables(python_executable))
        print("sharemyai_utils successfully installed.")

def copy_cuda_dlls(python_path):
    # After installing torch, copy the CUDA dlls to the python's directory
    print("Copying CUDA dlls to python's directory")
    torch_lib_path = python_path / 'Lib' / 'site-packages' / 'torch' / 'lib'
    print("Torch lib path:", torch_lib_path)
    for cu_file in torch_lib_path.glob("cu*64*.dll"):
        destination_file = python_path / cu_file.name
        if not destination_file.exists():
            try:
                shutil.copy(cu_file, python_path)
                print(f'Copied {cu_file} to {python_path}')
            except Exception as e:
                print(f'Failed to copy {cu_file} to {python_path}. Error: {str(e)}')
    for file in torch_lib_path.glob("nvrtc*.dll"):
        destination_file = python_path / file.name
        if not destination_file.exists():
            try:
                shutil.copy(file, python_path)
                print(f'Copied {file} to {python_path}')
            except Exception as e:
                print(f'Failed to copy {file} to {python_path}. Error: {str(e)}')
    for file in torch_lib_path.glob("zlibwapi.dll"):
        destination_file = python_path / file.name
        if not destination_file.exists():
            try:
                shutil.copy(file, python_path)
                print(f'Copied {file} to {python_path}')
            except Exception as e:
                print(f'Failed to copy {file} to {python_path}. Error: {str(e)}')

def set_environment_variables(python_path, settings=None):
    env = os.environ.copy()
    env['CUDA_PATH'] = os.path.dirname(python_path)
    # Set the environment variables to the python's directory
    env['PYTHONHOME'] = '' #os.path.dirname(python_executable)
    env['PYTHONPATH'] = '' #os.path.join(os.path.dirname(python_executable), 'Lib', 'site-packages')
    env['PYTHON_PATH'] = os.path.dirname(python_path)
    env['PYTHONEXECUTABLE'] = \
    env['PYTHON_EXECUTABLE'] = \
    env['PYTHON_BIN_PATH'] = os.path.join(os.path.dirname(python_path), 'python.exe')
    env['PYTHONWEXECUTABLE'] = \
    env['PYTHON_WEXECUTABLE'] = os.path.join(os.path.dirname(python_path), 'pythonw.exe')
    env['PYTHON_LIB_PATH'] = os.path.join(os.path.dirname(python_path), 'Lib', 'site-packages')
    env['PYTHON_SCRIPTS_PATH'] = os.path.join(os.path.dirname(python_path), 'Scripts')  # Added the Python Scripts folder to the environment variables
    env['VIRTUAL_ENV'] = os.path.dirname(python_path)
    # add our settings to the environment variables
    if settings:
        if 'model_cache_directory' in settings:
            # env['HF_HOME'] = settings['model_cache_directory']
            # store them in the plugin env so they get removed together with the plugin
            env['HF_HOME'] = os.path.join(os.path.dirname(python_path), 'hf_home')
            os.makedirs(env['HF_HOME'], exist_ok=True)
        if 'torch_device' in settings:
            env['TORCH_DEVICE'] = settings['torch_device']
        if 'training_directory' in settings:
            env['TRAINING_DIRECTORY'] = settings['training_directory']
        if 'inference_directory' in settings:
            env['INFERENCE_DIRECTORY'] = settings['inference_directory']
        if 'hf_token' in settings:
            env['HF_TOKEN'] = settings['hf_token']
    return env

def run_app(python_executable, script_path, settings=None):
    print("Running", script_path)
    print("Python executable:", python_executable)
    # Use subprocess.check_call instead of subprocess.Popen to launch the script and auto pipe the output
    subprocess.check_call([python_executable, script_path], env=set_environment_variables(python_executable, settings))

def main(plugin_path, settings):
    print('python started')
    print(os.path.join(plugin_path, 'model.json'))
    model = json.load(open(os.path.join(plugin_path, 'model.json')))
    requirements = model["requirements"]
    torch_requirements = model["torchRequirements"]
    env_name = model["name"]
    script_path = os.path.join(plugin_path, "run.py")
    python_ver = "3.10.6"
    venvs_path = settings["python_env_directory"] if settings and "python_env_directory" in settings else "./venvs"
    python_path = Path(venvs_path, env_name)
    python_path.mkdir(parents=True, exist_ok=True)
    install_python(python_ver, python_path)
    python_executable = python_path / 'python.exe'
    # install_uv(python_executable)
    print('installing torch dependencies')
    install_torch_dependencies(python_executable, torch_requirements)
    print('successfully installed torch dependencies')
    copy_cuda_dlls(python_path)
    print('installing dependencies')
    install_dependencies(python_executable, requirements)
    print('successfully installed dependencies')
    print('installing sharemyai_utils')
    # install_sharemyai_utils(python_executable)
    install_sharemyai(python_executable)
    print('successfully installed sharemyai_utils')
    run_app(python_executable, script_path, settings)

if __name__ == "__main__":
    plugin_path = sys.argv[1] if len(sys.argv) > 1 else "plugins/Code/sd-test"
    json_settings_path = sys.argv[2] if len(sys.argv) > 2 else None
    print("Plugin path:", plugin_path)
    print("Settings path:", json_settings_path)
    settings = json.loads(open(json_settings_path).read()) if json_settings_path else None
    main(plugin_path, settings)




__version__ = '0.2.1'
from .defaults import *
from .utils import *
from .images import *
from .videos import *
from .gpt4all import *
from .run_v2 import main as run_v2
from .run_v1 import main as run_v1




'''
run v2 plugins - gradio/web interfaces
'''
import os
import subprocess
import sys
import shutil
import json
import urllib.request
import zipfile
from pathlib import Path

def download_file(url, savepath):
    """
    Download the file and save to savepath
    """
    print(f'Downloading {url}')
    urllib.request.urlretrieve(url, savepath)

def install_python(python_ver, python_path):
    """
    Download and install python
    """
    URL_PYTHON = f'https://www.python.org/ftp/python/{python_ver}/python-{python_ver}-embed-amd64.zip'
    URL_PIP = r'https://bootstrap.pypa.io/get-pip.py'
    URL_MSVC = r'https://sharemyai-releases-dev.s3.amazonaws.com/msvc.zip'
    python_zip_path = python_path / 'python.zip'
    pip_path = python_path / 'get-pip.py'
    msvc_zip_path = python_path / 'msvc.zip'
    
    # Download files if they do not exist
    if not python_zip_path.exists() and not (python_path / 'python.exe').exists():
        download_file(URL_PYTHON, python_zip_path)
    if not pip_path.exists() and not (python_path / 'Scripts' / 'pip.exe').exists():
        download_file(URL_PIP, pip_path)
    if not msvc_zip_path.exists() and not any(python_path.glob('msvc*.dll')):
        download_file(URL_MSVC, msvc_zip_path)
    
    # Extract python zip if not already extracted
    if not (python_path / 'python.exe').exists():
        print("Extracting python")
        with zipfile.ZipFile(python_zip_path, 'r') as zip_ref:
            zip_ref.extractall(python_path)
        python_zip_path.unlink()
    
    # Remove _pth file if it exists
    for pth_file in python_path.glob("*._pth"):
        pth_file.unlink()
    
    # Install MS VC dlls if not already installed
    if msvc_zip_path.exists() and not (python_path / 'msvc').exists():
        print("Installing MSVC dlls")
        with zipfile.ZipFile(msvc_zip_path, 'r') as zip_ref:
            zip_ref.extractall(python_path)
        msvc_zip_path.unlink()
    
    # Install pip if not already installed
    if not (python_path / 'Scripts' / 'pip.exe').exists():
        print("pip not found, installing pip")
        subprocess.Popen(args='python.exe get-pip.py', cwd=str(python_path), shell=True).wait()
        pip_path.unlink()

    # Install dload if not already installed
    if not (python_path / 'Scripts' / 'dload.exe').exists():
        subprocess.Popen(args='python.exe -m pip install dload', cwd=str(python_path), shell=True).wait()
    
    # Install GitPython if not already installed
    if not (python_path / 'Scripts' / 'git.exe').exists():
        print("GitPython not found, installing GitPython")
        subprocess.Popen(args='python.exe -m pip install gitpython', cwd=str(python_path), shell=True).wait()

    # install gdown
    if not (python_path / 'Scripts' / 'gdown.exe').exists():
        subprocess.Popen(args='python.exe -m pip install gdown', cwd=str(python_path), shell=True).wait()

# https://astral.sh/blog/uv lets go fast
def install_uv(python_executable):
    subprocess.check_call([python_executable, "-m", "pip", "install", "uv"], env=set_environment_variables(python_executable))

def install_sharemyai(python_executable):
    subprocess.check_call([python_executable, "-m", "pip", "install", "sharemyai_utils", "--upgrade"], env=set_environment_variables(python_executable))


def install_dependencies(python_executable, plugin_full_path, dependencies, dependencies_overrides, dependencies_remove):
    filtered_dependencies = []
    pip_dependencies = []
    if "requirements.txt" in dependencies:
        requirements_path = f"{plugin_full_path}/requirements.txt"
        with open(requirements_path, "r") as req_file:
            all_requirements = req_file.readlines()
            for req in all_requirements:
                req = req.strip()
                if req and not req.startswith("#") and "torch" not in req and "xformers" not in req:
                    # Check for overrides
                    if req in dependencies_overrides:
                        req = dependencies_overrides[req]
                    if req in dependencies_remove:
                        continue
                    filtered_dependencies.append(req)
        
        for requirement in filtered_dependencies:
            subprocess.check_call([python_executable, "-m", "pip", "install", requirement], env=set_environment_variables(python_executable))
        dependencies.remove("requirements.txt")
    
    environment_file = "environment.yaml" if "environment.yaml" in dependencies else "environment.yml" if "environment.yml" in dependencies else None
    if environment_file:
        environment_path = f"{plugin_full_path}/{environment_file}"
        with open(environment_path, "r") as env_file:
            in_pip_section = False
            for line in env_file:
                if line.strip() == "pip:":
                    in_pip_section = True
                elif in_pip_section and line.startswith("  -"):
                    pip_dependency = line.strip()[2:]
                    if "torch" not in pip_dependency and "xformers" not in pip_dependency:
                        # Check for overrides
                        if pip_dependency in dependencies_overrides:
                            pip_dependency = dependencies_overrides[pip_dependency]
                        pip_dependencies.append(pip_dependency)
                elif in_pip_section and not line.startswith("  -"):
                    in_pip_section = False
        for pip_dependency in pip_dependencies:
            subprocess.check_call([python_executable, "-m", "pip", "install", pip_dependency], env=set_environment_variables(python_executable))
        dependencies.remove(environment_file)
    
    if dependencies:  # Check if there are any dependencies left to install after potentially removing "requirements.txt" and environment file
        remaining_dependencies = [dep for dep in dependencies] #if "torch" not in dep and "xformers" not in dep]
        if remaining_dependencies:
            subprocess.check_call([python_executable, "-m", "pip", "install"] + remaining_dependencies, env=set_environment_variables(python_executable))

def install_torch_dependencies(python_executable, dependencies):
    print("Python architecture:", "64-bit" if sys.maxsize > 2**32 else "32-bit")
    torch_index_url = "https://download.pytorch.org/whl/cu118"
    print("Installing torch dependencies from", torch_index_url)
    subprocess.check_call([python_executable, "-m", "pip", "install"] + dependencies + ["--index-url", torch_index_url], env=set_environment_variables(python_executable))
    
def install_sharemyai_utils(python_executable):
    # temp workaround for sharemyai_utils
    # Check and install specific version of setuptools if version changed
    setuptools_version = "58.2.0"
    try:
        installed_version = subprocess.check_output([python_executable, "-m", "pip", "show", "setuptools"], env=set_environment_variables(python_executable)).decode().split('\n')[1].split(': ')[1].strip()
        if installed_version != setuptools_version:
            print(f"Updating setuptools from version {installed_version} to {setuptools_version}")
            subprocess.check_call([python_executable, "-m", "pip", "install", f"setuptools=={setuptools_version}"], env=set_environment_variables(python_executable))
    except subprocess.CalledProcessError:
        print("setuptools not found, installing version", setuptools_version)
        subprocess.check_call([python_executable, "-m", "pip", "install", f"setuptools=={setuptools_version}"], env=set_environment_variables(python_executable))
    sharemyai_utils_path = os.path.join(os.path.dirname(__file__), "sharemyai_utils")
    sharemyai_utils_dest_path = os.path.join(os.path.dirname(python_executable), "sharemyai_utils")
    
    # Check if sharemyai_utils version has changed before attempting re-installation or copying
    try:
        installed_sharemyai_utils_version = subprocess.check_output([python_executable, "-m", "pip", "show", "sharemyai_utils"], env=set_environment_variables(python_executable)).decode().split('\n')[1].split(': ')[1]
        with open(os.path.join(sharemyai_utils_path, 'sharemyai_utils', '__init__.py'), 'r') as f:
            local_version = f.readline().strip().split('=')[1].replace("'", "").strip()
        installed_sharemyai_utils_version = installed_sharemyai_utils_version.strip()
        local_version = local_version.strip()
        print(f"Installed sharemyai_utils version: {installed_sharemyai_utils_version}, Local sharemyai_utils version: {local_version}")
        if installed_sharemyai_utils_version.strip() == local_version.strip():
            print("Versions match. No update required.")
        else:
            print("Version mismatch detected. Update required.")
            print(f"Updating sharemyai_utils from version {installed_sharemyai_utils_version} to {local_version}")
            print("Copying sharemyai_utils to", sharemyai_utils_dest_path)
            shutil.copytree(sharemyai_utils_path, sharemyai_utils_dest_path, dirs_exist_ok=True)
            subprocess.check_call([python_executable, "-m", "pip", "install", "-e", sharemyai_utils_dest_path], env=set_environment_variables(python_executable))
            print("sharemyai_utils successfully installed or updated.")
    except subprocess.CalledProcessError as e:
        print("Installing sharemyai_utils for the first time.")
        print("Copying sharemyai_utils to", sharemyai_utils_dest_path)
        shutil.copytree(sharemyai_utils_path, sharemyai_utils_dest_path, dirs_exist_ok=True)
        subprocess.check_call([python_executable, "-m", "pip", "install", "-e", sharemyai_utils_dest_path], env=set_environment_variables(python_executable))
        print("sharemyai_utils successfully installed.")

def copy_cuda_dlls(python_path):
    # After installing torch, copy the CUDA dlls to the python's directory
    print("Copying CUDA dlls to python's directory")
    torch_lib_path = python_path / 'Lib' / 'site-packages' / 'torch' / 'lib'
    print("Torch lib path:", torch_lib_path)
    for cu_file in torch_lib_path.glob("cu*64*.dll"):
        destination_file = python_path / cu_file.name
        if not destination_file.exists():
            try:
                shutil.copy(cu_file, python_path)
                print(f'Copied {cu_file} to {python_path}')
            except Exception as e:
                print(f'Failed to copy {cu_file} to {python_path}. Error: {str(e)}')
    for file in torch_lib_path.glob("nvrtc*.dll"):
        destination_file = python_path / file.name
        if not destination_file.exists():
            try:
                shutil.copy(file, python_path)
                print(f'Copied {file} to {python_path}')
            except Exception as e:
                print(f'Failed to copy {file} to {python_path}. Error: {str(e)}')
    for file in torch_lib_path.glob("zlibwapi.dll"):
        destination_file = python_path / file.name
        if not destination_file.exists():
            try:
                shutil.copy(file, python_path)
                print(f'Copied {file} to {python_path}')
            except Exception as e:
                print(f'Failed to copy {file} to {python_path}. Error: {str(e)}')

def set_environment_variables(python_path, settings=None):
    env = os.environ.copy()
    env['CUDA_PATH'] = os.path.dirname(python_path)
    # Set the environment variables to the python's directory
    env['PYTHONHOME'] = '' #os.path.dirname(python_executable)
    env['PYTHONPATH'] = '' #os.path.join(os.path.dirname(python_executable), 'Lib', 'site-packages')
    env['PYTHON_PATH'] = os.path.dirname(python_path)
    env['PYTHONEXECUTABLE'] = \
    env['PYTHON_EXECUTABLE'] = \
    env['PYTHON_BIN_PATH'] = os.path.join(os.path.dirname(python_path), 'python.exe')
    env['PYTHONWEXECUTABLE'] = \
    env['PYTHON_WEXECUTABLE'] = os.path.join(os.path.dirname(python_path), 'pythonw.exe')
    env['PYTHON_LIB_PATH'] = os.path.join(os.path.dirname(python_path), 'Lib', 'site-packages')
    env['PYTHON_SCRIPTS_PATH'] = os.path.join(os.path.dirname(python_path), 'Scripts')  # Added the Python Scripts folder to the environment variables
    env['VIRTUAL_ENV'] = os.path.dirname(python_path)
    # add our settings to the environment variables
    if settings:
        if 'model_cache_directory' in settings:
            env['HF_HOME'] = settings['model_cache_directory']
        if 'torch_device' in settings:
            env['TORCH_DEVICE'] = settings['torch_device']
        if 'training_directory' in settings:
            env['TRAINING_DIRECTORY'] = settings['training_directory']
        if 'inference_directory' in settings:
            env['INFERENCE_DIRECTORY'] = settings['inference_directory']
        if 'hf_token' in settings:
            env['HF_TOKEN'] = settings['hf_token']
    return env

def forward_output(pipe, label):
    for line in iter(pipe.readline, b''):
        print(f"{label}: {line.decode()}", end='')

def run_app(python_executable, plugin_path, run_commands, settings=None):
    print("Running", run_commands)
    print("Python executable:", python_executable)
    # attach our plugin path to the run command
    run_commands[0] = run_commands[0] #os.path.join(plugin_path, )
    # Use subprocess.Popen to launch the script with the updated working directory
    commands = [python_executable] + run_commands
    subprocess.check_call(commands, env=set_environment_variables(python_executable, settings), cwd=plugin_path)

def sanitize_path(path):
    return path.replace('\\x07', 'x07').replace('\\', '/')

def download_plugin(python_executable, url, plugin_path):
    """
    Download the plugin from the git url using a subprocess to utilize gitpython installed in the python executable path
    """
    plugin_path = sanitize_path(plugin_path)  # Sanitize the path
    print(f'Downloading {url} to {plugin_path}')
    # make the folder if it doesn't exist
    os.makedirs(plugin_path, exist_ok=True)
    download_command = f'from git import Repo; Repo.clone_from("{url}", "{plugin_path}")'
    subprocess.run([python_executable, "-c", download_command], check=True)

def run_postinstall(python_executable, postinstall_commands):
    for command in postinstall_commands:
        print("Running postinstall command:", command)
        try:
            subprocess.check_call([python_executable, "-c", command], env=set_environment_variables(python_executable))
        except subprocess.CalledProcessError as e:
            print(f"Error running postinstall command '{command}': {e}")
            sys.exit(1)

def run_runfile_overrides(runfile_path, runfile_overrides):
    # our runfile_overrides are key: value for string replacement (inline)
    # Open the python file to read its contents
    with open(runfile_path, 'r', encoding='utf-8') as file:
        code = file.read()
    # Replace each instance of key in our python file with our value, ensuring not to duplicate replacements
    for key, value in runfile_overrides.items():
        print(f"Replacing '{key}' with '{value}' in {runfile_path}")
        # Improved check to avoid duplicate replacements by examining the context around found strings
        start_index = 0
        while True:
            start_index = code.find(key, start_index)
            if start_index == -1:  # Key not found, move to next key
                break
            print(f"Replacing '{key}' with '{value}' in {runfile_path} at index {start_index}")
            # Check the surroundings of the found key to ensure it's not already replaced
            pre_context = code[max(0, start_index - len(value)):start_index]
            post_context = code[start_index + len(key):start_index + len(key) + len(value)]
            if value == "" or pre_context != value or post_context != value:  # If surroundings don't match the value, replace
                code = code[:start_index] + value + code[start_index + len(key):]
                start_index += len(value) if value else len(key)  # Move past the newly replaced segment or the key if value is empty
            else:
                start_index += len(key)  # Move past the current found key

    # Write the modified code back to the python file
    with open(runfile_path, 'w', encoding='utf-8') as file:
        file.write(code)
    print(f'Wrote runfile overrides to {runfile_path}')
            
def main(plugin_path, settings):
    pluginJson = json.load(open(plugin_path))
    plugin_name = pluginJson["name"]
    run_commands = pluginJson["run"]
    runfile = pluginJson.get("runfile", "")
    postinstall_commands = pluginJson.get("postInstall", [])
    runfile_overrides = pluginJson.get("runfileOverrides", {})
    python_ver = "3.10.6"
    venvs_path = settings["python_env_directory"] if settings and "python_env_directory" in settings else "./venvs"
    python_path = Path(venvs_path, plugin_name)
    python_path.mkdir(parents=True, exist_ok=True)
    install_python(python_ver, python_path)
    # Check if our plugin exists in the directory of the plugin JSON file, if not download it 
    plugin_directory = os.path.dirname(plugin_path)
    plugin_name = pluginJson["name"]
    python_executable = python_path / 'python.exe'
    # install_uv(python_executable)
    plugin_full_path = os.path.join(plugin_directory, plugin_name)
    if not os.path.exists(plugin_full_path) or not os.listdir(plugin_full_path):
        download_plugin(python_executable, pluginJson["remoteUrl"], plugin_full_path)
        # if we downloaded a plugin, check if it is sitting in a folder, if so move the contents to the plugin directory
        downloaded_plugin_path = plugin_full_path
        if len(os.listdir(downloaded_plugin_path)) == 1:
            inner_folder_path = os.path.join(downloaded_plugin_path, os.listdir(downloaded_plugin_path)[0])
            if os.path.isdir(inner_folder_path):  # Ensure it's a directory
                for file in os.listdir(inner_folder_path):
                    shutil.move(os.path.join(inner_folder_path, file), downloaded_plugin_path)
                os.rmdir(inner_folder_path)
    install_torch_dependencies(python_executable, pluginJson["torchRequirements"])
    copy_cuda_dlls(python_path)
    install_dependencies(python_executable, plugin_full_path, pluginJson["requirements"], pluginJson.get('requirementsOverrides', {}), pluginJson.get('requirementsRemove', []))
    # install_sharemyai_utils(python_executable)
    install_sharemyai(python_executable)
    # Change the current working directory to the plugin path so that relative paths in the Python app are based on the plugin path
    print("Changing current working directory to", plugin_full_path)
    # os.chdir(plugin_full_path)
    run_postinstall(python_executable=python_executable, postinstall_commands=postinstall_commands)
    if runfile_overrides and runfile != "":
        runfile_full_path = os.path.join(plugin_full_path, runfile)
        run_runfile_overrides(runfile_full_path, runfile_overrides)
    
    run_app(python_executable, plugin_full_path, run_commands, settings)

if __name__ == "__main__":
    plugin_path = sys.argv[1] if len(sys.argv) > 1 else ""
    json_settings_path = sys.argv[2] if len(sys.argv) > 2 else None
    print("Plugin path:", plugin_path)
    print("Settings path:", json_settings_path)
    settings = json.loads(open(json_settings_path).read()) if json_settings_path else None
    main(plugin_path, settings)




import asyncio
import base64
import time
from dotenv import load_dotenv, set_key
import os
import json
import psutil
import requests
import subprocess
import sys
import asyncio
from websockets.sync.client import connect

from cli.fingerprint import get_fingerprint
base_url = "https://sharemy.ai/api"
dotenv_path = '.env'
load_dotenv(dotenv_path)

def login_with_token(token):
    try:
        os.environ['SESSION_TOKEN'] = token
        # Check if SESSION_TOKEN already exists and update it; otherwise, add it
        if 'SESSION_TOKEN' in os.environ:
            # Update the SESSION_TOKEN value in the .env file
            with open(dotenv_path, 'r') as file:
                lines = file.readlines()
            with open(dotenv_path, 'w') as file:
                for line in lines:
                    if line.startswith('SESSION_TOKEN='):
                        file.write(f'SESSION_TOKEN={token}\n')
                    else:
                        file.write(line)
        else:
            # Append the SESSION_TOKEN to the .env file
            with open(dotenv_path, 'a') as file:
                file.write(f'SESSION_TOKEN={token}\n')

        # Load the updated environment variables
        load_dotenv(dotenv_path)

        user = get_current_user()
        if not user:
            # Remove the token from .env
            with open(dotenv_path, 'r') as file:
                lines = file.readlines()
            with open(dotenv_path, 'w') as file:
                for line in lines:
                    if not line.startswith('SESSION_TOKEN='):
                        file.write(line)
            print("Login failed. Please try again.")
            return False
        return user
    except Exception as e:
        print(f"An error occurred: {str(e)}")
        return False

def get_current_user():
    token = os.getenv('SESSION_TOKEN')
    if not token:
        print("No session token found. Please login first.")
        return
    response = requests.get(f"{base_url}/user", headers={"Authorization": f"Bearer {token}"})
    print('get_current_user response:', response)
    return response.json()

def get_user_points():
    token = os.getenv('SESSION_TOKEN')
    if not token:
        print("No session token found. Please login first.")
        return
    response = requests.get(f"{base_url}/user/points", headers={"Authorization": f"Bearer {token}"})
    return response.json()

def post_register_worker(pluginId=None):
    token = os.getenv('SESSION_TOKEN')
    if not token:
        print("No session token found. Please login first.")
        return
    fingerprint_bytes = asyncio.run(get_fingerprint())
    fingerprint_str = fingerprint_bytes.hex()
    data = {"fingerprint": fingerprint_str}
    if pluginId is not None:
        data["pluginId"] = pluginId
    response = requests.post(f"{base_url}/plugin/worker", headers={"Authorization": f"Bearer {token}"}, json=data)
    # handle a 401
    if response.status_code == 401:
        print('Unauthorized, login first')
        exit()
    print('post_register_worker response:', response)
    return response.json()
 
def get_my_workers():
    token = os.getenv('SESSION_TOKEN')
    if not token:
        print("No session token found. Please login first.")
        return
    response = requests.get(f"{base_url}/plugin/worker", headers={"Authorization": f"Bearer {token}"})
    return response.json()

def get_and_download_plugin(pluginId, base_path):
    token = os.getenv('SESSION_TOKEN')
    if not token:
        print("No session token found. Please login first.")
        return
    response = requests.get(f"{base_url}/plugin?id={pluginId}", headers={"Authorization": f"Bearer {token}"})
    remotePlugin = response.json()
    if not remotePlugin:
        raise Exception(f"Could not find plugin with id {pluginId}")
    
    pluginPath = os.path.join(base_path, remotePlugin['name'])
    if not os.path.exists(pluginPath):
        os.makedirs(pluginPath, exist_ok=True)
    
    modelJson = {
        "name": remotePlugin['name'],
        "description": remotePlugin['description'],
        "isPublic": remotePlugin['isPublic'],
        "remoteUrl": remotePlugin['remoteUrl'],
        "capabilities": remotePlugin['capabilities'],
        "requirements": remotePlugin['requirements'],
        "torchRequirements": remotePlugin['torchRequirements'],
        "params": remotePlugin['params'],
    }
    
    runPyPath = os.path.join(pluginPath, 'run.py')
    modelJsonPath = os.path.join(pluginPath, 'model.json')
  
    with open(runPyPath, 'w') as run_py:
        run_py.write(remotePlugin['code'])
    with open(modelJsonPath, 'w') as model_json:
        json.dump(modelJson, model_json, indent=2)
    
    return remotePlugin, pluginPath

def create_venv(plugin_path):
    venv_path = os.path.join(plugin_path, 'venv')
    if not os.path.exists(venv_path):
        subprocess.check_call([sys.executable, "-m", "venv", venv_path])
    return venv_path

def install_requirements(venv_path, requirements_list):
    pip_path = 'Scripts' if os.name == 'nt' else 'bin'
    subprocess.check_call([os.path.join(venv_path, pip_path, 'pip'), 'install', '--upgrade', "sharemyai-utils"])
    subprocess.check_call([os.path.join(venv_path, pip_path, 'pip'), 'install', *requirements_list])

def install_torch_requirements(venv_path, dependencies):
    pip_path = 'Scripts' if os.name == 'nt' else 'bin'
    torch_index_url = "https://download.pytorch.org/whl/cu118"
    print("Installing torch dependencies from", torch_index_url)
    subprocess.check_call([os.path.join(venv_path, pip_path, 'pip'), 'install'] + dependencies + ["--index-url", torch_index_url])
    
def run_plugin(venv_path, plugin_path, plugin_name):
    # make ./models if it doesnt exist
    models_path = os.path.join('.', 'models')
    os.makedirs(models_path, exist_ok=True)
    # Adjust plugin_name to create a valid directory path
    plugin_config_path = os.path.join(models_path, plugin_name.replace('/', os.sep))
    os.makedirs(plugin_config_path, exist_ok=True)
    # make the server config json 
    with open(os.path.join(plugin_config_path, "server_config.json"), 'w') as server_config_json:
        json.dump({}, server_config_json)
    os.environ['HF_HOME'] = './models'
    pip_path = 'Scripts' if os.name == 'nt' else 'bin'
    process = subprocess.Popen([os.path.join(venv_path, pip_path, 'python'), os.path.join(plugin_path, 'run.py')])
    print('Running plugin', plugin_name, 'with pid', process.pid)
    return process.pid

def get_inference_job(worker_id):
    token = os.getenv('SESSION_TOKEN')
    if not token:
        print("No session token found. Please login first.")
        return
    response = requests.get(f"{base_url}/inference/request/process?workerId={worker_id}", headers={"Authorization": f"Bearer {token}"})
    return response.json()

def process_inference_job(prompt, input_image, params, model_params, job_id):
    token = os.getenv('SESSION_TOKEN')
    if not token:
        print("No session token found. Please login first.")
        return
    params = {**model_params, **params}
    params.pop('image', None)
    params.pop('prompt', None)

    body = {
        'prompt': prompt,
        **params,
    }
    if input_image:
        if isinstance(input_image, str):
            body['image'] = input_image
        else:
            # Assuming input_image is a file-like object
            body['image'] = base64.b64encode(input_image.read()).decode('utf-8')
   
    try:
        with connect("ws://localhost:3030/ws/generate") as websocket:
            websocket.send(json.dumps(body))
            while True:
                message = websocket.recv()
                json_message = json.loads(message)
                if "error" in json_message:
                    print(f"Error: {json_message['error']}")
                    result_response = requests.delete(f"{base_url}/inference/result/submit", headers={"Authorization": f"Bearer {token}"}, json={
                                  "resultId": job_id,
                                  "errorMessage": f"Error processing inference job: {json_message['error']}",
                                })
                    print('result_response:', result_response.json())
                    break
                elif "message" in json_message:
                    print(f"Message: {json_message['message']}")
                elif "progress" in json_message:
                    print(f"Progress: {json_message['progress']}%, Step: {json_message.get('step', '?')}, Total Steps: {json_message.get('inference_steps', '?')}")
                elif "file_paths" in json_message or "file_path" in json_message:
                    formatted_paths = json_message['file_paths'] if "file_paths" in json_message else [json_message['file_path']]
                    print("Generated Image Paths:", formatted_paths)
                    for path in json_message['file_paths'] if "file_paths" in json_message else [json_message['file_path']]:
                        filename = os.path.basename(path)
                        signed_response = requests.get(f"{base_url}/presigned?file={filename}", headers={"Authorization": f"Bearer {token}"})
                        presigned_url = signed_response.json().get('presignedUrl')
                        image_path = signed_response.json().get('imagePath')
                        
                        # Determine the content type based on the file extension
                        _, file_extension = os.path.splitext(filename)
                        content_type = {
                            '.png': 'image/png',
                            '.jpg': 'image/jpeg',
                            '.jpeg': 'image/jpeg',
                            '.gif': 'image/gif',
                            '.bmp': 'image/bmp',
                            '.tiff': 'image/tiff',
                            '.webp': 'image/webp',
                            '.svg': 'image/svg+xml',
                        }.get(file_extension.lower(), 'application/octet-stream')
                        
                        # Upload the image to the presigned URL
                        with open(path, 'rb') as f:
                            upload_response = requests.put(presigned_url, data=f.read(), headers={'Content-Type': content_type})
                            if upload_response.status_code == 200:
                                result_response = requests.post(f"{base_url}/inference/result/submit", headers={"Authorization": f"Bearer {token}"}, json={
                                    "resultId": job_id,
                                    "originalFilename": filename,
                                    "output": image_path,
                                })
                                print('Upload success')
                                # sleep for 1s between jobs
                                time.sleep(1)
                                break
                            else:
                                print(f"Failed to upload {filename}. Status code: {upload_response.status_code}")
                                result_response = requests.delete(f"{base_url}/inference/result/submit", headers={"Authorization": f"Bearer {token}"}, json={
                                    "resultId": job_id,
                                    "errorMessage": f"Error processing inference job - Upload {filename} to {image_path} - {upload_response.status_code}",
                                })
                        pass
                    break
                else:
                    print(json_message)
                    print("Unhandled message type received.")
    except Exception as e:
        print('No websocket connection:', e)



def kill_process_on_port(port):
    for proc in psutil.process_iter(['pid', 'name', 'cmdline']):
        try:
            # Check if the process is listening on the specified port
            for conn in proc.connections(kind='inet'):
                if conn.laddr.port == port:
                    print(f"Found process {proc.pid} - {proc.info['name']} listening on port {port}")
                    print(f"Command line: {' '.join(proc.info['cmdline'])}")
                    
                    # Terminate the process
                    proc.terminate()
                    print(f"Process {proc.pid} terminated")
                    return
        except (psutil.NoSuchProcess, psutil.AccessDenied, psutil.ZombieProcess):
            pass
    
    print(f"No process found listening on port {port}")